/*
	code file for PROforma
	definition of a path aid commands

	(c) PROGS, Professional & Graphical Software
	written by Joachim Van der Auwera
	started 21 July 1993
	11.11.1996 support for save/restore of clipping path as part of Gstate
	29.11.1996 faster (and shorter) addflatcurve
	30.11.1996 PathNoClip now really doesn't have a clipping path
	08.01.1997 addflatcurve now replaced by a version which is not recursive,
			   and uses a fixed amount of stackspace (approx. 450 bytes). This
			   allows a curve to be broken down in maximum 2048 lines.
	15.06.1997 reintroduced proper test in "addflatcurve" routine
	04.08.1997 isflat could still cause divzero
*/

#include "gstate.h"
#include "system.h"
#include "list.h"
#include "path.h"

/* get a segment, if there are no segments left, then a new segment block */
/* will be allocated - returns zero when out of memory */
static segment *get_segment(Gstate gstate)
{
	segment *s;
	Error err;

	err=MEMUserAllocate(gstate->freesegments,&s);
	if (err) return NULL;
	return s;
}

/* command to clear the current path */
Error PFPathClear(Gstate gstate)
{
	segment *prev, *path, *next, *clip;

	path=gstate->path2;
	while (path)
	{
		next=path->next;
		free_segment(gstate,path);
		path=next;
	}
	gstate->path2=NULL;

	/* clear the current path */
	path=gstate->path;

	while (path)
	{
		next=path->next;
		free_segment(gstate,path);
		path=next;
	}

	/* clear the current point */
	gstate->subpathorg=gstate->posdev;

	/* copy the clipping path to the current path */
	/* this involves making a copy of all elements */
	path=get_segment(gstate);		/* allocate the first segment */
	if (!path) return ERR_IMEM;
	clip=gstate->clippath;
	MEMCopy((long)sizeof(segment),clip,path);
	path->prev=path->next=NULL;
	gstate->path=path;
	prev=path;
	clip=clip->clip;

	while(clip)
	{
		path=get_segment(gstate);	/* allocate the segment */
		if (!path) return ERR_IMEM;
		MEMCopy(sizeof(segment),clip,path);
		path->next=NULL;
		prev->next=path;
		path->prev=prev;
		prev=path;
		clip=clip->clip;
	}
	gstate->path->prev=prev;	  /* link with prev should be circular */
	gstate->last=prev;			  /* start value for linking */

	return ERR_OK;
}

/* command to clear the current path with an empty clipping path */
Error PathNoClip(Gstate gstate)
{
	segment *path, *next, *sh, *ss;

	/* release the current path */
	for (path=gstate->path ; path ; path=next)
	{
		next=path->next;
		free_segment(gstate,path);
	}

	/* allocate segments for clipping path */
	sh=get_segment(gstate); if (!sh) return ERR_IMEM;
	ss=get_segment(gstate); if (!ss) return ERR_IMEM;

	/* link as path */
	gstate->path=sh;
	gstate->last=ss;
	sh->next=sh->prev=ss;
	ss->next=NULL;
	ss->prev=sh;

	sh->starty=(short)0x8000; ss->starty=(short)0x7fff;   /* init boundaries */
	sh->startx=ptmin+pt_one; ss->startx=ptmax-pt_one;

	/* release the alternate path */
	for (path=gstate->path2 ; path ; path=next)
	{
		next=path->next;
		free_segment(gstate,path);
	}

	/* allocate segments for clipping path */
	sh=get_segment(gstate); if (!sh) return ERR_IMEM;
	ss=get_segment(gstate); if (!ss) return ERR_IMEM;

	/* link as path */
	gstate->path2=sh;
	gstate->last2=ss;
	sh->next=sh->prev=ss;
	ss->next=NULL;
	ss->prev=sh;

	sh->starty=(short)0x8000; ss->starty=(short)0x7fff;   /* init boundaries */
	sh->startx=ptmin+pt_one; ss->startx=ptmax-pt_one;

	return ERR_OK;
}

/* command to clear the clipping path */
Error PFClipClear(Gstate gstate)
{
	segment *sh, *ss, *s1, *s2;
	segment *next, *seg;

	/* clear old clipping path */
	seg=gstate->clippath;
	while (seg)
	{
		next=seg->clip;
		free_segment(gstate,seg);
		seg=next;
	}

	/* allocate segments for clipping path */
	sh=get_segment(gstate); if (!sh) return ERR_IMEM;
	ss=get_segment(gstate); if (!ss) return ERR_IMEM;
	s1=get_segment(gstate); if (!s1) return ERR_IMEM;
	s2=get_segment(gstate); if (!s2) return ERR_IMEM;

	/* link as clipping path */
	sh->clip=s1; s1->clip=s2; s2->clip=ss; ss->clip=NULL;

	sh->starty=(short)0x8000; ss->starty=(short)0x7fff;   /* init boundaries */
	sh->startx=ptmin+pt_one; ss->startx=ptmax-pt_one;
	s1->starty=s2->starty=(short)0x8001; s1->endy=s2->endy=(short)0x7ffe;
	s1->incr=s2->incr=0;
	s1->startx=ptmin+pt_one+pt_one; s2->startx=ptmax-pt_one-pt_one;
	s1->direction=PF_LINE_UP; s2->direction=PF_LINE_DOWN;
	sh->isclip=TRUE; s1->isclip=TRUE; s2->isclip=TRUE; ss->isclip=TRUE;

	gstate->clippath=sh;		  /* set empty clipping path */

	return PFPathClear(gstate);   /* clear path */
}

/* add a flattened line to the current path (device coordinates) */
Error addflatline(Gstate gstate, pt fromx, pt fromy, pt tox, pt toy)
{
	segment *seg, *path;
	short sy, ey, keepy;
	pt keepx, incr, adjustment;

	sy=pt2rshort(fromy); ey=pt2rshort(toy);

	if (sy==ey) return ERR_OK;			/* horizontal line -> OK */

	seg=get_segment(gstate);
	if (!seg) return ERR_IMEM;

	/* make sure the direction is correct */
	seg->direction=PF_LINE_DOWN;
	if (sy>ey)
	{
		seg->direction=PF_LINE_UP;
		keepy=ey; ey=sy; sy=keepy;
		keepx=tox; tox=fromx; fromx=keepx;
		keepx=toy; toy=fromy; fromy=keepx;
	}

	/* fill in the rest of the segment */
	incr=fixdiv((tox-fromx),(toy-fromy));
	adjustment=fixmul(incr,fromy-short2pt(sy));

	/* the adjustment is there because just rounding the y coordinates may
	   change the position of the line drasticaly and may change from one side
	   of a pixel centre to the other (or even several pixel centres). */

	seg->starty=sy; seg->endy=ey;
	seg->startx=fromx-adjustment+incr/2;
	seg->incr=incr;
	seg->isclip=FALSE;

	/* now link this item into the path */
	path=gstate->last;
	/* find item before which we have to link in */
	while(path->starty>=sy) path=path->prev;
	while(path->starty<sy) path=path->next;
	/* link in before segment *path */
	seg->prev=path->prev;
	seg->next=path;
	seg->prev->next=seg;
	path->prev=seg;

	gstate->last=seg;

	return ERR_OK;
}

Error addflatline2(Gstate gstate, pt fromx, pt fromy, pt tox, pt toy)
{
	segment *seg, *path;
	short sy, ey, keepy;
	pt keepx, incr, adjustment;

	if (!gstate->path2) return ERR_OK;

	sy=pt2rshort(fromy); ey=pt2rshort(toy);

	if (sy==ey) return ERR_OK;			/* horizontal line -> OK */

	seg=get_segment(gstate);
	if (!seg) return ERR_IMEM;

	/* make sure the direction is correct */
	seg->direction=PF_LINE_DOWN;
	if (sy>ey)
	{
		seg->direction=PF_LINE_UP;
		keepy=ey; ey=sy; sy=keepy;
		keepx=tox; tox=fromx; fromx=keepx;
		keepx=toy; toy=fromy; fromy=keepx;
	}

	/* fill in the rest of the segment */
	incr=fixdiv((tox-fromx),(toy-fromy));
	adjustment=fixmul(incr,fromy-short2pt(sy));

	/* the adjustment is there because just rounding the y coordinates may
	   change the position of the line drasticaly and may change from one side
	   of a pixel centre to the other (or even several pixel centres). */

	seg->starty=sy; seg->endy=ey;
	seg->startx=fromx-adjustment+incr/2;
	seg->incr=incr;
	seg->isclip=FALSE;

	/* now link this item into the path */
	path=gstate->last2;
	/* find item before which we have to link in */
	while(path->starty>=sy) path=path->prev;
	while(path->starty<sy) path=path->next;
	/* link in before segment *path */
	seg->prev=path->prev;
	seg->next=path;
	seg->prev->next=seg;
	path->prev=seg;

	gstate->last2=seg;

	return ERR_OK;
}

/* Routine to flatten a curve. */
/* Recursive routine, using tail recursion */
/* needs device coordinates */
#ifdef RECURVE_FLATTEN_CURVE
Error addflatcurve(Gstate gstate, pt x0, pt y0, pt x1, pt y1,
				   pt x2, pt y2, pt x3, pt y3, pt flatness,
				   Error (*line)(), Bitmap bitmap)
{
	Error err;
	pt a,b;
	pt x01, x12, x02;
	pt y01, y12, y02;

#define midpoint(a,b) ((a + b)/2)
top:
	/* use subdivision to approximate the curve using a polygon */
	/* we check whether the required accuracy (or "flatness") is reached */
	/* by checking whether the distance from either curvepoint to */
	/* the midpoint between the controlpoints is smaller than the flatness */
	/* in both directions. This should be much faster than the real test. */

	/* nothing further subdividing when no vertical pixel difference */
	if (pt2rshort(y0)==pt2rshort(y3) &&
		pt2rshort(y0)==pt2rshort(y2) &&
		pt2rshort(y0)==pt2rshort(y1))
		return line(bitmap, x0, y0, x3, y3);

	x12=midpoint(x1,x2);
	y12=midpoint(y1,y2);

	if (abs(x0-x12)<=flatness && abs(x3-x12)<=flatness &&
		abs(y0-y12)<=flatness && abs(y3-y12)<=flatness)
	{
		/* good enough, approximate using a line */
		return line(bitmap, x0, y0, x3, y3);
	}
	else
	{
		x01=midpoint(x0,x1);
		y01=midpoint(y0,y1);
		x02=midpoint(x01,x12);
		y02=midpoint(y01,y12);

		/* Update coordinates for the second half. */
		x2=midpoint(x2,x3);
		y2=midpoint(y2,y3);
		x1=midpoint(x12,x2);
		y1=midpoint(y12,y2);
		a =midpoint(x02,x1);
		b =midpoint(y02,y1);
		catch( addflatcurve(gstate, x0,y0, x01,y01, x02,y02, a,b,
							flatness, line, bitmap) );
		x0=a; y0=b;
		goto top;
	}
	return ERR_OK;
}
#else
static int isflat(pt dx, pt dy, pt x1, pt y1, pt x2, pt y2, pt flatness)
{
	pt t,d;
	if (abs(dx)>abs(dy))
	{
		t=fixdiv(dy,dx);
		d=fixmul(fixsqrt(pt_one+fixmul(t,t)),flatness);
		if (abs(y1-fixmul(x1,t)) <= d)
			return TRUE;
		if (abs(y2-fixmul(x2,t)) <= d)
			return TRUE;
	}
	else
	{
		t = dy!=0 ? fixdiv(dx,dy) : 0;
		d=fixmul(fixsqrt(pt_one+fixmul(t,t)),flatness);
		if (abs(x1-fixmul(y1,t)) <= d)
			return TRUE;
		if (abs(x2-fixmul(y2,t)) <= d)
			return TRUE;
	}
	return FALSE;
}

Error addflatcurve(Gstate gstate, pt x0, pt y0, pt x1, pt y1,
				   pt x2, pt y2, pt x3, pt y3, pt flatness,
				   Error (*line)(), Bitmap bitmap)
{
	typedef struct { pt x0,y0,x1,y1,x2,y2,x3,y3; } curvepoints;
#define MAX_NESTING 12
	int nesting=1;
	curvepoints curve[MAX_NESTING];
	curvepoints *next=&curve[1];
	Error err;
	pt x12, y12;

#define midpoint(a,b) ((a + b)/2)
	while (nesting)
	{
		/* use subdivision to approximate the curve using a polygon */
		/* we check whether the required accuracy (or "flatness") is reached */
		/* by checking whether the distance from either curvepoint to */
		/* the line between the endpoints <= flatness. */
		/* The flatness requirement is met when abd(H)<=sqrt(L)*flatness */
		/* where H=y1-(x1*t), L=1+t*t, t=y3/x3 */
		/* coordinates are normalised to have the origin at x0,y0 */
		/* and another time for x2,y2 instead of x1,y1 */
		/* when dx<dy (abs) then x and y have to be swapped */

		/* no subdividing when - no vertical pixel difference */
		/*					OR - max nesting reached */
		/*					OR - flatness reached */
		if (nesting==MAX_NESTING ||
			(pt2rshort(y0)==pt2rshort(y3) &&
			 pt2rshort(y0)==pt2rshort(y2) &&
			 pt2rshort(y0)==pt2rshort(y1) ) ||
			isflat(x3-x0, y3-y0, x1-x0, y1-y0, x2-x0, y2-y0, flatness) )
		{
			catch( line(bitmap, x0, y0, x3, y3) );
			nesting--; next--;
			x0=next->x0; y0=next->y0; x1=next->x1; y1=next->y1;
			x2=next->x2; y2=next->y2; x3=next->x3; y3=next->y3;
		}
		else
		{
			/* subdivide */
			x12=midpoint(x1,x2);
			y12=midpoint(y1,y2);
			next->x3=x3;
			next->y3=y3;
			next->x2=midpoint(x2,x3);
			next->y2=midpoint(y2,y3);
			next->x1=midpoint(x12,next->x2);
			next->y1=midpoint(y12,next->y2);
			x1=midpoint(x0,x1);
			y1=midpoint(y0,y1);
			x2=midpoint(x1,x12);
			y2=midpoint(y1,y12);
			x3=next->x0=midpoint(x2,next->x1);
			y3=next->y0=midpoint(y2,next->y1);
			nesting++; next++;
		}
	}
	return ERR_OK;
}
#endif RECURVE_FLATTEN_CURVE

/* here come the routine to set the clipping path */
/* this involves a routine to add a segment to a clipping path */
/* and a routine to create an empty clipping path with no segments */

/* command to release clipping path */
Error ReleaseClipPath(Gstate gstate)
{
	segment *seg, *next;

	/* clear old clipping path */
	seg=gstate->clippath;
	while (seg)
	{
		next=seg->clip;
		free_segment(gstate,seg);
		seg=next;
	}
	return ERR_OK;
}

/* command to clear the clipping path */
static Error NoClipPath(Gstate gstate)
{
	segment *sh, *ss;

	if (gstate->clippathmine) ReleaseClipPath(gstate);
	gstate->clippathmine=TRUE;

	/* allocate segments for clipping path */
	sh=get_segment(gstate); if (!sh) return ERR_IMEM;
	ss=get_segment(gstate); if (!ss) return ERR_IMEM;

	/* link as clipping path */
	sh->clip=ss; ss->clip=NULL;

	sh->starty=(short)0x8000; ss->starty=(short)0x7fff;   /* init boundaries */
	sh->startx=ptmin+pt_one; ss->startx=ptmax-pt_one;
	sh->isclip=TRUE; ss->isclip=TRUE;

	gstate->clippath=sh;		  /* set empty clipping path */
}

/* add a flattened line to the current clipping path (device coordinates) */
static Error addclipline(Gstate gstate, short sy, short ey, pt
						 fromx, pt incr, char dir)
{
	segment *seg, *path;

	if (sy==ey) return ERR_OK;			/* horizontal line -> OK */

	seg=get_segment(gstate);
	if (!seg) return ERR_IMEM;

	/* fill in the segment */
	seg->direction=dir;
	seg->starty=sy; seg->endy=ey;
	seg->startx=fromx+incr/2;
	seg->incr=incr;
	seg->isclip=TRUE;

	/* now link this item into the path */
	path=gstate->clippath;
	/* find item after which we have to link in */
	while(path->clip->starty<sy) path=path->clip;
	/* link in after segment *path */
	seg->clip=path->clip;
	path->clip=seg;

	return ERR_OK;
}

/* routine to add a trapezoid to the clipping path */
static Error clip_trapezoid(Bitmap gstate, short starty, short endy,
					pt leftstartx, pt leftincr, pt rightstartx, pt rightincr)
{
	addclipline((Gstate)gstate, starty, endy, leftstartx , leftincr , PF_LINE_UP);
	addclipline((Gstate)gstate, starty, endy, rightstartx, rightincr, PF_LINE_DOWN);
	return ERR_OK;
}


Error PFPathClip(Gstate gstate)
{
	short rule=PF_WINDING_RULE;

	if (gstate->pathtype==PF_PATH_EOFILLED) rule=PF_EVEN_ODD_RULE;

	NoClipPath(gstate);
	fillpath(gstate, clip_trapezoid, (Bitmap)gstate, rule);
	return PFPathClear(gstate);
}
